/*

  sshgenvalist.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 1999 SSH Communications Security, Finland
  All rights reserved.

  Created Mon Jun 12 08:20:16 2000.

  */

/* WARNING: This file was generated using genvalist-data.pl. The 
   contents of this file should not be modified, instead make the 
   required modifications to files:

     genvalist-data.pl
       The Perl program which combines the pieces and generated this 
       file. 

     genvalist-data.txt
       A plain text file which contains extra include file descriptions 
       and supported type listing. 

     genvalist-raw.c
       File that contains the actual C implementation of required 
       functions. Part of this file is left unspecified, as the 
       Perl program fills these gaps. This file is used to generate
       the file sshgenvalist.c.

     genvalist-raw.h
       The raw header file. The Perl program uses this to generate 
       sshgenvalist.h. 
*/

#include "sshincludes.h"
#include "sshgenvalist.h"

#define SSH_DEBUG_MODULE "SshGenVaList"

/* Generating hash table. */
/* Table of supported types (indexed using the hash function which 
   follows below. */    

#define SSH_GEN_VALIST_HASH_INIT_VALUE ((unsigned int)1575574585)
#define SSH_GEN_VALIST_HASH_TABLE_SIZE ((unsigned int)50)

static SshGenVaListType 
ssh_gen_valist_type_table[SSH_GEN_VALIST_HASH_TABLE_SIZE] = 
{
  0,
  0,
  0,
  SSH_GEN_VALIST_unsigned_char_pp,
  0,
  0,
  SSH_GEN_VALIST_unsigned_int,
  0,
  SSH_GEN_VALIST_SshBuffer,
  0,
  SSH_GEN_VALIST_SshGenVaListEnum_p,
  0,
  0,
  SSH_GEN_VALIST_char,
  0,
  SSH_GEN_VALIST_SshMPInt_p,
  0,
  SSH_GEN_VALIST_int_p,
  0,
  0,
  SSH_GEN_VALIST_unsigned_char,
  SSH_GEN_VALIST_unsigned_char_p,
  0,
  0,
  0,
  SSH_GEN_VALIST_SshGenVaListEnum_pp,
  0,
  0,
  SSH_GEN_VALIST_Boolean_p,
  SSH_GEN_VALIST_void_pp,
  SSH_GEN_VALIST_SshMPInt,
  0,
  SSH_GEN_VALIST_Boolean_pp,
  0,
  0,
  SSH_GEN_VALIST_int_pp,
  SSH_GEN_VALIST_char_p,
  SSH_GEN_VALIST_SshBuffer_p,
  0,
  SSH_GEN_VALIST_Boolean,
  SSH_GEN_VALIST_unsigned_int_p,
  SSH_GEN_VALIST_void_p,
  0,
  SSH_GEN_VALIST_SshGenVaListEnum,
  SSH_GEN_VALIST_int,
  SSH_GEN_VALIST_unsigned_int_pp,
  0,
  0,
  0,
  SSH_GEN_VALIST_char_pp
};

void ssh_gen_valist_realloc(SshGenVaList vl, unsigned int mem)
{
  if (vl->max < mem)
    {
      vl->max = mem * 2;
      vl->entry = ssh_xrealloc(vl->entry,
                               sizeof(SshGenVaListEntries) * vl->max);
    }
}

SshGenVaList ssh_gen_valist_alloc(SshGenVaListDef def,
                                  unsigned int def_num_elem,
                                  va_list ap)
{
  SshGenVaList vl;
  SshGenVaListEnum fmt;

  vl      = ssh_xcalloc(1, sizeof(*vl));

  while (1)
    {
      size_t i;
      unsigned int hash, t, counter;
      const unsigned char *types;
      
      fmt = va_arg(ap, SshGenVaListEnum);

      /* Get the type string. Here we would like to use faster methods
         to seek for the match. However, it is unreasonable to assume that
         the implementor would actually make sure that the input is
         in suitable order. (Although we could spend O(n log n) steps
         as precomputation to sort them, or O(n) to throw these into
         a hash table.) */
      for (i = 0; i < def_num_elem; i++)
        if (def[i].tag == fmt)
          {
            break;
          }
      SSH_ASSERT(i < def_num_elem);
      
      types = def[i].types;

      /* Make an entry for the enum type itself. */
      ssh_gen_valist_realloc(vl, vl->used+1);
      vl->entry[vl->used].c.data_SshGenVaListEnum = fmt;
      vl->entry[vl->used].type = SSH_GEN_VALIST_SshGenVaListEnum;
      vl->used++;
      
      /* Parse and read. */
      for (i = 0, hash = SSH_GEN_VALIST_HASH_INIT_VALUE, counter = 0; 1; i++)
        {
          if (types[i] == ',' ||
              types[i] == '\0')
            {
              /* Force the hash size to be suitable. */
              hash %= SSH_GEN_VALIST_HASH_TABLE_SIZE;

              /* Assert that enough space is available. */
              ssh_gen_valist_realloc(vl, vl->used+1);
              
              /* The vararg reading portion. */
              
              switch (ssh_gen_valist_type_table[hash])
                {
                case SSH_GEN_VALIST_END:
                  break;
                case SSH_GEN_VALIST_Boolean:
                  vl->entry[vl->used].c.data_Boolean =
                  va_arg(ap, Boolean);
                  break;
                case SSH_GEN_VALIST_Boolean_p:
                  vl->entry[vl->used].c.data_Boolean_p =
                  va_arg(ap, Boolean *);
                  break;
                case SSH_GEN_VALIST_Boolean_pp:
                  vl->entry[vl->used].c.data_Boolean_pp =
                  va_arg(ap, Boolean **);
                  break;
                case SSH_GEN_VALIST_unsigned_char:
                  vl->entry[vl->used].c.data_unsigned_char =
                  va_arg(ap, unsigned int);
                  break;
                case SSH_GEN_VALIST_unsigned_char_p:
                  vl->entry[vl->used].c.data_unsigned_char_p =
                  va_arg(ap, unsigned char *);
                  break;
                case SSH_GEN_VALIST_unsigned_char_pp:
                  vl->entry[vl->used].c.data_unsigned_char_pp =
                  va_arg(ap, unsigned char **);
                  break;
                case SSH_GEN_VALIST_char:
                  vl->entry[vl->used].c.data_char =
                  va_arg(ap, int);
                  break;
                case SSH_GEN_VALIST_char_p:
                  vl->entry[vl->used].c.data_char_p =
                  va_arg(ap, char *);
                  break;
                case SSH_GEN_VALIST_char_pp:
                  vl->entry[vl->used].c.data_char_pp =
                  va_arg(ap, char **);
                  break;
                case SSH_GEN_VALIST_unsigned_int:
                  vl->entry[vl->used].c.data_unsigned_int =
                  va_arg(ap, unsigned int);
                  break;
                case SSH_GEN_VALIST_unsigned_int_p:
                  vl->entry[vl->used].c.data_unsigned_int_p =
                  va_arg(ap, unsigned int *);
                  break;
                case SSH_GEN_VALIST_unsigned_int_pp:
                  vl->entry[vl->used].c.data_unsigned_int_pp =
                  va_arg(ap, unsigned int **);
                  break;
                case SSH_GEN_VALIST_int:
                  vl->entry[vl->used].c.data_int =
                  va_arg(ap, int);
                  break;
                case SSH_GEN_VALIST_int_p:
                  vl->entry[vl->used].c.data_int_p =
                  va_arg(ap, int *);
                  break;
                case SSH_GEN_VALIST_int_pp:
                  vl->entry[vl->used].c.data_int_pp =
                  va_arg(ap, int **);
                  break;
                case SSH_GEN_VALIST_void_p:
                  vl->entry[vl->used].c.data_void_p =
                  va_arg(ap, void *);
                  break;
                case SSH_GEN_VALIST_void_pp:
                  vl->entry[vl->used].c.data_void_pp =
                  va_arg(ap, void **);
                  break;
                case SSH_GEN_VALIST_SshMPInt:
                  vl->entry[vl->used].c.data_SshMPInt =
                  va_arg(ap, SshMPInt);
                  break;
                case SSH_GEN_VALIST_SshMPInt_p:
                  vl->entry[vl->used].c.data_SshMPInt_p =
                  va_arg(ap, SshMPInt *);
                  break;
                case SSH_GEN_VALIST_SshBuffer:
                  vl->entry[vl->used].c.data_SshBuffer =
                  va_arg(ap, SshBuffer);
                  break;
                case SSH_GEN_VALIST_SshBuffer_p:
                  vl->entry[vl->used].c.data_SshBuffer_p =
                  va_arg(ap, SshBuffer *);
                  break;
                case SSH_GEN_VALIST_SshGenVaListEnum:
                  vl->entry[vl->used].c.data_SshGenVaListEnum =
                  va_arg(ap, SshGenVaListEnum);
                  break;
                case SSH_GEN_VALIST_SshGenVaListEnum_p:
                  vl->entry[vl->used].c.data_SshGenVaListEnum_p =
                  va_arg(ap, SshGenVaListEnum *);
                  break;
                case SSH_GEN_VALIST_SshGenVaListEnum_pp:
                  vl->entry[vl->used].c.data_SshGenVaListEnum_pp =
                  va_arg(ap, SshGenVaListEnum **);
                  break;
                default:
                  ssh_fatal("ssh_gen_valist_alloc: invalid type requested.");
                  break;
                }

              vl->entry[vl->used].type =
                ssh_gen_valist_type_table[hash];
              vl->used++;

              /* Check if we can quit now. */
              if (types[i] == '\0')
                break;
              
              /* Restart. */
              hash = SSH_GEN_VALIST_HASH_INIT_VALUE;
              counter = 0;
              continue;
            }

          /* Hash code. */
          if (types[i] == ' ')
            continue;

          t = types[i] + counter;
          if (hash & 1)
            hash = (hash >> 1) ^ ((unsigned int)0xedb88320);
          else
            hash = (hash >> 1);
          hash ^= t;
          counter++;
        }

      /* Check for the next. */
    }
  
  return vl;
}

void ssh_gen_valist_free(SshGenVaList vl)
{
  memset(vl->entry, 0, sizeof(struct SshGenVaListEntryRec)*vl->max);
  ssh_xfree(vl->entry);
  memset(vl, 0, sizeof(SshGenVaList));
  ssh_xfree(vl);
}

/* The library defines SSH_GEN_VALIST_REWIND_START and
   SSH_GEN_VALIST_REWIND_END for the caller to use instead of
   explicit values.

   If you are using more than, say, 64 THOUSAND arguments to a C
   function you are really cool! ...and in that case you SHOULD
   get into trouble anyway.
   */

void ssh_gen_valist_rewind(SshGenVaList vl, int steps)
{
  int newpos;

  newpos = vl->pos + steps;
  if (newpos < 0)
    newpos = 0;
  if (newpos >= vl->used)
    newpos = vl->used-1;

  vl->pos = newpos;
}

/* genvalist.c */








